/* VMSOBJ_XABALL.C - 07-FEB-1999 ZE.
   Allocation XAB (eXtended Access Block) object. 

   24-MAR-1999 ZE. -- replace some PyErr_SetString() with PyErr_Format()
*/

/* ------------------------------------------------------------------------- */

#include "Python.h"
#include "vmsdef.h"

#include <xaballdef.h>
#include "vmsobj__def.h"
#include "vmsobj__xabtst.h"	/* test for _any_ XAB */

/* ------------------------------------------------------------------------- */
/* Generic routine to convert binary data to a Python type. */

extern PyObject * vms__cvt_bin2py
        (char                    * ab_data_addr         /* data address */
        ,unsigned short int        w_data_type          /* data type	*/
        ,unsigned short int        w_bufsiz             /* data length	*/
        );
/* ------------------------------------------------------------------------- */
/* Generic routine to convert a Python type to binary data. */
extern long vms__cvt_py2bin
        (PyObject                * ar_py_data           /* Python data  */
        ,char                    * ab_data_addr         /* data address */
        ,unsigned short int        w_data_type          /* data type    */
        ,unsigned short int        w_bufsiz             /* data length  */
        );
/* ------------------------------------------------------------------------- */

PyTypeObject vmsobj_xaball_type;

/* ------------------------------------------------------------------------- */
vmsobj_xaball *
vmsobj_xaball_new (struct XABALL *arg)
{
	vmsobj_xaball *xp;

	/* -------------------- */
	/* new 'vmsobj_xaball' object */
	xp = PyObject_NEW (vmsobj_xaball, &vmsobj_xaball_type);
	if (xp == NULL)
	{
	    return NULL;
	}
	xp->l_flags = 0;
	/* pointers to other objects that the XABALL points to */
	xp->ar_nxt  = NULL;	/* next XAB object */

	/* -------------------- */
	if (arg == NULL)
	{
	    /* allocate memory for real XABALL and note it's address */
	    xp->ab_xaball = malloc (XAB$K_ALLLEN);
	    if (xp->ab_xaball == NULL)
	    {
		PyMem_DEL(xp);
		(void) PyErr_NoMemory();
		return NULL;
	    }
	    *xp->ab_xaball  = cc$rms_xaball;
	    xp->l_flags |= VMSOBJ_M_ALLOC;
	}
	else
	{
	    /* 'arg' has address of already existing XABALL */
	    xp->ab_xaball = arg;
	    /* VMSOBJ_M_ALLOC is clear */
	}

	/* -------------------- */
	return xp;
}

/* ------------------------------------------------------------------------- */
/* vmsobj_xaball methods */

static void
vo_xaball_dealloc (vmsobj_xaball *xp)
{
	/* was XABALL allocated on object creation? */
	if (xp->l_flags & VMSOBJ_M_ALLOC)
	{
	    /* release memory of VMS XABALL */
	    if (xp->ab_xaball != NULL)
	    {
		(void) free (xp->ab_xaball);
	    }
	}

	/* -------------------- */
	/* delete references to other objects the XABALL might point to */
	if (xp->ar_nxt != NULL)
	{
	    Py_DECREF(xp->ar_nxt);	/* next XAB object */
	}

	/* -------------------- */
	/* release memory of object */
	PyMem_DEL(xp);
}

/* ------------------------------------------------------------------------- */
static PyMethodDef vo_xaball_methods[] = {
	{NULL,		NULL}		/* sentinel */
};

/* ------------------------------------------------------------------------- */

struct vmsobj_xaball_attr {
	char			* at_attrnam;	/* name of attribute */
	short int		  w_offset;	/* offset within XABALL */
	unsigned short int	  w_typ;	/* data type	     */
	long			  l_mask;	/* bit address	     */
};
typedef struct vmsobj_xaball_attr vmsobj_xaball_attr;

/* XAB$B_COD + XAB$B_BLN are not changeable as they identify a XABALL */

#define ROAT	0x8000	/* read-only attribute */
#define OBIO	0x4000	/* object I/O */
#define OFSMSK	0x1fff	/* mask byte offset count */

#define OB_K_NXT 1 /* next XAB object */

static vmsobj_xaball_attr vmsobj_r_xaball_attr [] = {
	{"B_AID",	23,	ITMTYP_K_BYTE,0	} ,
	{"B_ALN",	9,	ITMTYP_K_BYTE,0	} ,
	{"L_ALQ",	16,	ITMTYP_K_LONG,0	} ,
	{"B_AOP",	8,	ITMTYP_K_BYTE,0	} ,
	 {"M_HRD",	8,	ITMTYP_K_WORD,XAB$M_HRD } ,
	 {"M_ONC",	8,	ITMTYP_K_WORD,XAB$M_ONC } ,
	 {"M_CBT",	8,	ITMTYP_K_WORD,XAB$M_CBT } ,
	 {"M_CTG",	8,	ITMTYP_K_WORD,XAB$M_CTG } ,
	{"B_BKZ",	22,	ITMTYP_K_BYTE,0	} ,

	{"B_BLN",	ROAT+1,	ITMTYP_K_BYTE,0	} ,
	{"B_COD",	ROAT+0,	ITMTYP_K_BYTE,0	} ,
	{"W_DEQ",	20,	ITMTYP_K_WORD,0	} ,
	{"L_LOC",	12,	ITMTYP_K_LONG,0	} ,
	{"NXT",		OBIO+OB_K_NXT,
				ITMTYP_K_BOOL,0	} , /* object I/O */
	{"L_NXT",	ROAT+4,	ITMTYP_K_LONG,0	} ,
	{"W_RFI",	24,	ITMTYP_K_WORD,0	} , /* 3W - special code */
	 {"W_RFI0",	24,	ITMTYP_K_WORD,0	} ,
	 {"W_RFI2",	26,	ITMTYP_K_WORD,0	} ,
	 {"W_RFI4",	28,	ITMTYP_K_WORD,0	} ,
	{"W_VOL",	10,	ITMTYP_K_WORD,0	} ,
	{NULL,		0,	0,	0	}
};

/* ------------------------------------------------------------------------- */

static PyObject *
vo_xaball_getattr (vmsobj_xaball *xp, char *name)
{
	unsigned long l_attr_idx;

	l_attr_idx = 0;
	while (vmsobj_r_xaball_attr[l_attr_idx].at_attrnam != NULL)
	{
	  if (strcmp (name, vmsobj_r_xaball_attr[l_attr_idx].at_attrnam) == 0)
	  {
	    /* size of item types - in PYVMS_GLOBAL.C */
	    extern long vmsdef_gl_itmtypsiz[];
	    char * ab_data;

	    ab_data = (char *)xp->ab_xaball;
	    /* ignore ROAT for get attribute */
	    /* if (vmsobj_r_xaball_attr[l_attr_idx].w_offset >= 0) */

	    if (vmsobj_r_xaball_attr[l_attr_idx].w_offset & OBIO)	    
	    {
	      /* object I/O */
	      unsigned short int  w_objtyp;

	      w_objtyp = vmsobj_r_xaball_attr[l_attr_idx].w_offset & OFSMSK;
	      switch (w_objtyp)
	      {
	      /* ---------- */
	      case OB_K_NXT:
		/* return with a Python object, if one was assigned before */
		if (xp->ar_nxt != NULL)
		{
		    Py_INCREF (xp->ar_nxt);
		    return xp->ar_nxt;
		}
		Py_INCREF (Py_None);
		return Py_None;
	      /* ---------- */
	      default:
		return PyErr_Format(PyExc_SystemError,
			"BUGCHK - object I/O - unsupported object type: %d",
			(unsigned int)w_objtyp);
	      }
	    } /* object I/O */
	    else
	    { /* BWL,M  I/O */
	      ab_data += (vmsobj_r_xaball_attr[l_attr_idx].w_offset & OFSMSK);
	      if (vmsobj_r_xaball_attr[l_attr_idx].l_mask == 0)
	      {
		/* special check for W_RFI, which is 3W */
		if (
		    ((vmsobj_r_xaball_attr[l_attr_idx].w_offset & OFSMSK) == 24) &&
		    (vmsobj_r_xaball_attr[l_attr_idx].w_typ == ITMTYP_K_WORD   )
		   )
		{
		  /* return resultant-value (3W) */
		  return vms__cvt_bin2py
		    (ab_data				     /* data address */
		    ,vmsobj_r_xaball_attr[l_attr_idx].w_typ  /* data type    */
		    ,(unsigned short int) 6		     /* data length  */
		  );
		}
		else
		{
		  /* return resultant-value */
		  return vms__cvt_bin2py
		    (ab_data				 /* data address */
		    ,vmsobj_r_xaball_attr[l_attr_idx].w_typ /* data type	 */
		    ,vmsdef_gl_itmtypsiz[		 /* data length	 */
			vmsobj_r_xaball_attr[l_attr_idx].w_typ
					]
		  );
		}
	      }
	      else
	      {
		/* return a BIT */
		if ((*ab_data) & vmsobj_r_xaball_attr[l_attr_idx].l_mask)
		{
		  return PyInt_FromLong( (long)1);
		}
		else
		{
		  return PyInt_FromLong( (long)0);
		}
	      }
	    } /* object or BLW,M I/O */
	  } /* attribute check */
	  l_attr_idx++;
	} /* while() - attribute loop */

	/* -------------------- */
	/* fallback */
	return Py_FindMethod(vo_xaball_methods, (PyObject *)xp, name);
} /* vo_xaball_getattr (vmsobj_xaball *xp, char *name) */

/* ------------------------------------------------------------------------- */

static int
vo_xaball_setattr (vmsobj_xaball *xp, char *name, PyObject *v)
{
	unsigned long l_attr_idx;

	l_attr_idx = 0;
	while (vmsobj_r_xaball_attr[l_attr_idx].at_attrnam != NULL)
	{
	  if (strcmp (name, vmsobj_r_xaball_attr[l_attr_idx].at_attrnam) == 0)
	  {
	    /* size of item types - in PYVMS_GLOBAL.C */
	    extern long   vmsdef_gl_itmtypsiz[];
	    long	  l_status;
	    char	* ab_data;
	    PyObject	* ar_v;

	    ab_data = (char *)xp->ab_xaball;

	    if ( !(vmsobj_r_xaball_attr[l_attr_idx].w_offset & ROAT))
	    {
	      if (vmsobj_r_xaball_attr[l_attr_idx].w_offset & OBIO)	    
	      {
		/* object I/O */
		unsigned short int   w_objtyp;

		w_objtyp = vmsobj_r_xaball_attr[l_attr_idx].w_offset & OFSMSK;
	        switch (w_objtyp)
	        {
		/* ---------- */
		case OB_K_NXT:
		  if (v == Py_None)
		  {
		    /* 'None' = break connection to next XAB */
		    if (xp->ar_nxt != NULL)
		    {
		      Py_DECREF(xp->ar_nxt);  /* drop old reference */
		      xp->ar_nxt = NULL;
		    }
		    xp->ab_xaball->xab$l_nxt = 0;
		    return 0; /* OK */
		  }
		  /* check and make sure this is _a_ XABxxx */
		  if (! vmsobj__anyxab_Check(v))
		  {
		    PyErr_SetString(PyExc_TypeError,
			"must be a XABxxx object or None");
		    return -1;
		  }
		  if (xp->ar_nxt != NULL)
		  {
		    Py_DECREF(xp->ar_nxt);  /* drop old reference */
		  }
		  xp->ar_nxt = v;		/* remember for Py_DECREF */
		  /* keep reference, because vmsobj_xaball points to object */
		  Py_INCREF (v);
		  /* A different XAB than XABALL might have been passed. */
		  /* All XAB objects must share the VMS' XAB address in  */
		  /* their structures. */
		  /* (char*) to keep DEC-C on Alpha quiet */
		  xp->ab_xaball->xab$l_nxt =
					(char *)((vmsobj_xaball *)v)->ab_xaball;
		  return 0; /* OK */
		/* ---------- */
		default:
		  (void) PyErr_Format(PyExc_SystemError,
			"BUGCHK - object I/O - unsupported object type: %d",
			(unsigned int)w_objtyp);
		  return -1;
		}
	      } /* object I/O */
	      else
	      { /* BWL,M  I/O */
		ab_data += (vmsobj_r_xaball_attr[l_attr_idx].w_offset & OFSMSK);
		if (vmsobj_r_xaball_attr[l_attr_idx].l_mask == 0)
		{
/* @@ check that BW don't overflow */
		 /* special check for W_RFI, which is 3W */
		 if (
		    ((vmsobj_r_xaball_attr[l_attr_idx].w_offset & OFSMSK) == 24) &&
		    (vmsobj_r_xaball_attr[l_attr_idx].w_typ == ITMTYP_K_WORD   )
		   )
		 {
		  /* convert Python type (3W) */
                  l_status = vms__cvt_py2bin
                    (v					 /* Python data  */
		    ,ab_data				 /* data address */
                    ,vmsobj_r_xaball_attr[l_attr_idx].w_typ /* data type	*/
                    ,(unsigned short int) 6		 /* data length	*/
		  );
		 }
		 else
		 {
                  /* convert Python type */
                  l_status = vms__cvt_py2bin
                    (v					 /* Python data  */
		    ,ab_data				 /* data address */
                    ,vmsobj_r_xaball_attr[l_attr_idx].w_typ /* data type	*/
                    ,vmsdef_gl_itmtypsiz[		 /* data length	*/
		     vmsobj_r_xaball_attr[l_attr_idx].w_typ
				      ]
		  );
		 }
		/* .. */
		}
		else
		{
		  long			l_v;
		  unsigned long		l_msk;
		  unsigned short int	w_msk;
		  unsigned char		b_msk;

		  /* set/clear a BIT */
		  if (!PyInt_Check(v))
		  {
		    PyErr_SetString(PyExc_AttributeError,
			"bitmask assignment needs integer");
		    return -1;
		  }
		  l_v = PyInt_AsLong(v);
		  if (PyErr_Occurred())
		  {
		    return -1;
		  }

		  if (l_v == 0)
		  {
		    /* clear bit */
		    l_msk = ~ vmsobj_r_xaball_attr[l_attr_idx].l_mask;
		    w_msk = l_msk;
		    b_msk = w_msk;

		    if (vmsobj_r_xaball_attr[l_attr_idx].w_typ == ITMTYP_K_BYTE)
		    {
		      *ab_data &= b_msk;
		    }
		    else
		    {
		      if (vmsobj_r_xaball_attr[l_attr_idx].w_typ == ITMTYP_K_WORD)
		      {
		        *ab_data &= w_msk;
		      }
		      else
		      {
			if (vmsobj_r_xaball_attr[l_attr_idx].w_typ == ITMTYP_K_LONG)
			{
			  *ab_data &= l_msk;
			}
			else
			{
			  /* @@ bad table */
			} /* -ITMTYP_K_LONG */
		      } /* -ITMTYP_K_WORD */
		    } /* -ITMTYP_K_BYTE */
		  }
		  else
		  {
		    /* set bit */
		    l_msk = vmsobj_r_xaball_attr[l_attr_idx].l_mask;
		    w_msk = l_msk;
		    b_msk = w_msk;

		    if (vmsobj_r_xaball_attr[l_attr_idx].w_typ == ITMTYP_K_BYTE)
		    {
		      *ab_data |= b_msk;
		    }
		    else
		    {
		      if (vmsobj_r_xaball_attr[l_attr_idx].w_typ == ITMTYP_K_WORD)
		      {
		        *ab_data |= w_msk;
		      }
		      else
		      {
			if (vmsobj_r_xaball_attr[l_attr_idx].w_typ == ITMTYP_K_LONG)
			{
			  *ab_data |= l_msk;
			}
			else
			{
			  /* @@ bad table */
			} /* -ITMTYP_K_LONG */
		      } /* -ITMTYP_K_WORD */
		    } /* -ITMTYP_K_BYTE */
		  } /* clear or set bit */
	          l_status = 1; /* OK */
	        } /* others or bitmask */
	      } /* object or BWL,M I/O */
	    } /* offset 0x8000 indicates readonly attribute */
	    else
	    {
		PyErr_SetString(PyExc_AttributeError,
		    "read-only vmsobj_xaball attribute");
		return -1;
	    }
	    if (l_status == 1)
	    {
	      l_status = 0;
	    }
	    return l_status;       /* possible error mesage has been set */
	  }
	  l_attr_idx++;
	}

	/* -------------------- */
	PyErr_SetString(PyExc_AttributeError,
	    "non-existing vmsobj_xaball attribute");
	return -1;
} /* static int vo_xaball_setattr (vmsobj_xaball *xp, char *name, PyObject *v) */

/* ------------------------------------------------------------------------- */

static PyObject *
vo_xaball_repr (vmsobj_xaball *ar_xaball)
{
	char buf[40];

	sprintf(buf, "<vmsobj_xaball, XABALL at 0x%08x>", (ar_xaball->ab_xaball));
	return PyString_FromString(buf);
} /* static PyObject * vo_xaball_repr (vmsobj_xaball *ar_xaball) */

/* ------------------------------------------------------------------------- */
PyTypeObject vmsobj_xaball_type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,				/*ob_size*/
	"vmsobj_xaball",		/*tp_name*/
	sizeof(vmsobj_xaball),		/*tp_basicsize*/
	0,				/*tp_itemsize*/
	/* methods */
	(destructor)vo_xaball_dealloc,	/*tp_dealloc*/
	(printfunc)0,			/*tp_print*/
	(getattrfunc)vo_xaball_getattr,	/*tp_getattr*/
	(setattrfunc)vo_xaball_setattr,	/*tp_setattr*/
	(cmpfunc)0,			/*tp_compare*/
	(reprfunc)vo_xaball_repr,	/*tp_repr*/
	0,				/*tp_as_number*/
	0,				/*tp_as_sequence*/
	0,				/*tp_as_mapping*/
	(hashfunc)0,			/*tp_hash*/
};

/* ------------------------------------------------------------------------- */

/* EOF: VMSOBJ_XABALL.C */
